from django.conf import settings
from django.db import models
from django.utils.functional import curry

from . import exceptions


def get_mappings():
    '''
    Get all mappings from the settings

    To use the Django Tags Input module the `TAGS_INPUT_SETTINGS` must be
    defined.
    '''
    return getattr(settings, 'TAGS_INPUT_MAPPINGS', {})


def get_mapping(model_or_queryset):
    '''Get the mapping for a given model or queryset'''
    mappings = get_mappings()

    if isinstance(model_or_queryset, models.query.QuerySet):
        queryset = model_or_queryset
        model = model_or_queryset.model
    elif issubclass(model_or_queryset, models.Model):
        queryset = model_or_queryset.objects.all()
        model = model_or_queryset
    else:
        raise TypeError(
            'Only `django.db.model.Model` and `django.db.query.QuerySet` '
            'objects are valid arguments')

    meta = model._meta
    mapping_key = meta.app_label + '.' + meta.object_name

    mapping = mappings.get(mapping_key)
    if mapping is not None:
        mapping = mapping.copy()
    else:
        raise exceptions.MappingUndefined('Unable to find mapping '
                                          'for %s' % mapping_key)

    # The callable allows for customizing the queryset on the fly
    queryset = mapping.get('queryset', queryset)
    if callable(queryset):
        queryset = queryset(mapping)

    mapping['app'] = meta.app_label
    mapping['model'] = meta.object_name
    mapping['queryset'] = queryset
    mapping.setdefault('separator', ' - ')

    if 'field' in mapping:
        mapping['fields'] = mapping['field'],
    elif 'fields' not in mapping:
        raise exceptions.ConfigurationError(
            'Every mapping should have a field or fields attribute. Mapping: '
            '%r' % mapping)

    mapping.setdefault('split_func', curry(
        mapping.get('split_func', split_func),
        mapping['fields'],
        mapping['separator'],
    ))
    mapping.setdefault('join_func', curry(
        mapping.get('join_func', join_func),
        mapping['fields'],
        mapping['separator'],
    ))
    mapping.setdefault('filter_func', curry(
        mapping.get('filter_func', filter_func),
        mapping['fields'],
        mapping['separator'],
    ))

    return mapping.copy()


def filter_func(fields, separator, values):
    filters = {}
    if values:
        values = [v.split(separator, len(fields)) for v in values]
        items = zip(fields, zip(*values))
        for field, value in items:
            filters['%s__in' % field] = value

    return filters


def join_func(fields, separator, values):
    return values['pk'], separator.join(values[field] for field in fields)


def split_func(fields, separator, value):
    return dict(zip(fields, value.split(separator, len(fields))))




try:  # pragma: no cover
    from django.apps import apps

    def get_model(app, model):
        apps_config = apps.get_app_config(app)
        return apps_config.get_model(model)

except ImportError:  # pragma: no cover
    from django.db import models

    def get_model(app, model):
        return models.get_model(app, model)

from django import http

try:  # pragma: no cover
    from django.utils import simplejson
except ImportError:  # pragma: no cover
    import json as simplejson

from . import utils


def _filter_func(queryset, field, term):
    return queryset.filter(**{'%s__istartswith' % field: term})


def autocomplete(request, app, model, fields):
    model = get_model(app, model)
    mapping = utils.get_mapping(model)
    fields = fields.split('-')

    raw_queryset = (
        mapping['queryset']
        .filter(**mapping.get('filters', {}))
        .exclude(**mapping.get('excludes', {}))
        .values('pk', *fields)
        .order_by(*mapping.get('ordering', fields))
    )
    autocomplete_filter = mapping.get('autocomplete_queryset_filter',
                                      _filter_func)
    term = request.GET.get('term')
    if term:
        queryset = mapping['queryset'].none()
        for field in fields:
            queryset |= autocomplete_filter(raw_queryset, field, term)
    else:
        queryset = raw_queryset

    max_results = request.GET.get('max_results')
    if max_results and max_results.isdigit():
        max_results = int(max_results)
    else:
        max_results = 10

    results = [mapping['join_func'](v)[1] for v in queryset[:max_results]]
    if results:
        response = simplejson.dumps(results),
    else:
        response = ''

    return http.HttpResponse(response, content_type='application/javascript')




__package_name__ = 'django-tags-input'
__version__ = '4.2.0'
__author__ = 'Rick van Hattem'
__author_email__ = 'wolph@wol.ph'
__description__ = ' '.join('''
Django jQuery Tags Input is a Django field and widget wrapper that adds an easy
to use interface for `ManyToManyField` and `ForeignKey` fields in the forms of
autocompleting tags with optionally live creation of new linked
items
'''.strip().split())
__url__ = 'https://github.com/WoLpH/django-tags-input'



import six

from django.conf import settings
from django import forms
from django.template.loader import render_to_string
from django import urls
from django.contrib.admin import widgets

try:  # pragma: no cover
    from collections import OrderedDict
except ImportError:  # pragma: no cover
    from django.utils.datastructures import SortedDict as OrderedDict


class TagsInputWidgetBase(forms.SelectMultiple):

    def __init__(self, on_add_tag=None, on_remove_tag=None, on_change_tag=None,
                 *args, **kwargs):
        self.on_add_tag = on_add_tag
        self.on_remove_tag = on_remove_tag
        self.on_change_tag = on_change_tag
        forms.SelectMultiple.__init__(self, *args, **kwargs)

    def build_attrs(self, base_attrs, extra_attrs=None, **kwargs):
        '''Compatibility function for the behavior changes in Django 1.11+'''
        attrs = dict(base_attrs or {}, **kwargs)
        if extra_attrs:  # pragma: no cover
            attrs.update(extra_attrs)
        return attrs

    def render(self, name, value, attrs=None, choices=(), renderer=None):
        context = self.build_attrs(attrs, name=name)
        context['on_add_tag'] = self.on_add_tag
        context['on_remove_tag'] = self.on_remove_tag
        context['on_change_tag'] = self.on_change_tag

        context['STATIC_URL'] = settings.STATIC_URL
        context['mapping'] = self.mapping
        context['autocomplete_url'] = urls.reverse(
            'tags_input:autocomplete',
            kwargs=dict(
                app=self.mapping['app'],
                model=self.mapping['model'],
                fields='-'.join(self.mapping['fields']),
            ),
        )

        if value:
            fields = self.mapping['fields']
            join_func = self.mapping['join_func']

            ids = []
            for v in value:
                if isinstance(v, six.integer_types):
                    ids.append(v)

            values_map = OrderedDict(map(
                join_func,
                self.mapping['queryset']
                .filter(pk__in=ids)
                .values('pk', *fields)
                .order_by('pk')
            ))

            values = []
            for v in value:
                if isinstance(v, six.integer_types):
                    values.append(values_map[v])
                else:
                    values.append(v)

            context['values'] = ', '.join(values)

        return render_to_string('tags_input_widget.html', context)

    def value_from_datadict(self, data, files, name):
        tags = data.get(name, '').split(',')
        incomplete = data.get(name + '_incomplete', '')
        if incomplete != data.get(name + '_default'):
            tags += incomplete.split(',')
        return [t.strip() for t in tags if t]


class TagsInputWidget(TagsInputWidgetBase):

    class Media:

        css = {
            'all': (
                'jquery-tagsinput-revisited-2.0/dist/'
                'jquery.tagsinput-revisited.min.css',
            ),
        }
        js = (
            'jquery-tagsinput-revisited-2.0/dist/'
            'jquery.tagsinput-revisited.min.js',
        )
        enable_jquery = getattr(settings, 'TAGS_INPUT_INCLUDE_JQUERY', True)
        if enable_jquery:  # pragma: no cover
            css['all'] += 'jquery-ui-1.12.1/jquery-ui.min.css',
            js = (
                'jquery-3.2.1.min.js',
                'jquery-ui-1.12.1/jquery-ui.min.js',
            ) + js


class AdminTagsInputWidget(
        widgets.FilteredSelectMultiple, TagsInputWidgetBase):
    # This class inherits FilteredSelectMultiple because the Django admin
    # handles the FilteredSelectMultiple differently from regular widgets

    @property
    def media(self):
        return forms.Media(js=self.Media.js, css=self.Media.css)

    class Media:

        css = getattr(settings, 'TAGS_INPUT_ADMIN_CSS', {
            'all': (
                'jquery-tagsinput-revisited-2.0/dist/'
                'jquery.tagsinput-revisited.min.css',
                'jquery-ui-1.12.1/jquery-ui.min.css',
            ),
        })
        js = getattr(settings, 'TAGS_INPUT_ADMIN_JS', (
            'jquery-3.2.1.min.js',
            'jquery-ui-1.12.1/jquery-ui.min.js',
            'jquery-tagsinput-revisited-2.0/dist/'
            'jquery.tagsinput-revisited.min.js',
        ))




from django.conf import urls
from . import views

app_name = 'tags_input'
urlpatterns = [
    urls.url(r'^autocomplete/(?P<app>\w+)/(?P<model>\w+)/(?P<fields>[\w-]+)/$',
             views.autocomplete, name='autocomplete'),
]




from . import fields
from . import widgets

from django.contrib import admin


class TagsInputMixin(object):
    def get_tag_fields(self):
        '''Get a list fo fields on this model that could be potentially tagged.

        By default reads self.tag_fields if it exists of returns None for
        default behavious.
        '''

        return getattr(self, 'tag_fields', None)

    def formfield_for_manytomany(self, db_field, request=None, **kwargs):
        '''
        Get a form Field for a ManyToManyField.
        '''
        # If it uses an intermediary model that isn't auto created, don't show
        # a field in admin.

        # if not db_field.rel.through._meta.auto_created:
        if not db_field.remote_field.through._meta.auto_created:
            return None

        # If there is a list of taggable fields, and this filed isn't one of
        # them, then fall back to parent method.
        tag_fields = self.get_tag_fields()

        if tag_fields and db_field.name not in tag_fields:
            print('in fields', tag_fields, db_field.name)
            return super(TagsInputMixin, self).formfield_for_manytomany(
                db_field, request, **kwargs)
        else:
            print('nin fields', tag_fields, db_field.name)

        queryset = db_field.related_model._default_manager.get_queryset()
        # queryset = db_field.rel.to._default_manager.get_queryset()

        kwargs['queryset'] = queryset
        kwargs['widget'] = widgets.AdminTagsInputWidget(
            verbose_name=db_field.verbose_name,
            is_stacked=db_field.name in self.filter_vertical,
            attrs=kwargs.get('attrs'),
            choices=kwargs.get('choices', ()), )
        kwargs['required'] = not db_field.blank
        kwargs['help_text'] = getattr(db_field, 'help_text', None)
        kwargs['verbose_name'] = getattr(db_field, 'verbose_name', None)

        # Ugly hack to stop the Django admin from adding the + icon

        if db_field.name not in self.raw_id_fields:
            self.raw_id_fields = list(self.raw_id_fields)
            self.raw_id_fields.append(db_field.name)

        return fields.AdminTagsInputField(**kwargs)


class TagsInputAdmin(TagsInputMixin, admin.ModelAdmin):
    pass


class TagsInputTabularInline(TagsInputMixin, admin.TabularInline):
    pass


class TagsInputStackedInline(TagsInputMixin, admin.StackedInline):
    pass



from django import forms
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _

from . import widgets
from . import utils


class TagsInputField(forms.ModelMultipleChoiceField):
    widget = widgets.TagsInputWidget
    default_error_messages = {
        'list': _('Enter a list of values.'),
        'invalid_choice': _('Select a valid choice. %s is not one of the'
                            ' available choices.'),
        'invalid_pk_value': _('"%s" is not a valid value for a primary key.')
    }

    def __init__(self, queryset, **kwargs):
        self.create_missing = kwargs.pop('create_missing', False)
        self.mapping = kwargs.pop('mapping', None)
        super(TagsInputField, self).__init__(queryset, **kwargs)
        self.widget.mapping = self.get_mapping()

    def get_mapping(self):
        if not self.mapping:
            self.mapping = mapping = utils.get_mapping(self.queryset)
            mapping['queryset'] = self.queryset
            mapping['create_missing'] = (
                self.create_missing or mapping.get('create_missing', False)
            )

        return self.mapping

    def clean(self, value):
        mapping = self.get_mapping()
        fields = mapping['fields']
        filter_func = mapping['filter_func']
        join_func = mapping['join_func']
        split_func = mapping['split_func']

        values = dict(
            join_func(v)[::-1] for v in self.queryset
            .filter(**filter_func(value))
            .values('pk', *fields)
        )
        values = dict((k.lower(), v) for k, v in values.items())
        missing = [v for v in value if v.lower() not in values]
        if missing:
            if mapping['create_missing']:
                for v in value:
                    if v in missing:
                        o = self.queryset.model(**split_func(v))
                        o.clean()
                        o.save()
                        values[v.lower()] = o.pk
            else:
                raise ValidationError(
                    self.error_messages['invalid_choice'],
                    code='invalid_choice',
                    params={'value': ', '.join(missing)},
                )

        ids = []
        for v in value:
            ids.append(values[v.lower()])

        return forms.ModelMultipleChoiceField.clean(self, ids)


class AdminTagsInputField(TagsInputField):
    widget = widgets.AdminTagsInputWidget

    def __init__(self, queryset, verbose_name=None, *args, **kwargs):
        TagsInputField.__init__(self, queryset, *args, **kwargs)

        if verbose_name:  # pragma: no branch
            self.label = verbose_name




import os
import sys
import setuptools
from setuptools.command.test import test as TestCommand


# To prevent importing about and thereby breaking the coverage info we use this
# exec hack
about = {}
with open('tags_input/__about__.py') as fp:
    exec(fp.read(), about)


if os.path.isfile('README.rst'):
    long_description = open('README.rst').read()
else:
    long_description = ('See http://pypi.python.org/pypi/' +
                        about['__package_name__'])


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.test_args)
        sys.exit(errno)


if __name__ == '__main__':
    setuptools.setup(
        name=about['__package_name__'],
        version=about['__version__'],
        author=about['__author__'],
        author_email=about['__author_email__'],
        description=about['__description__'],
        url=about['__url__'],
        license='BSD',
        packages=setuptools.find_packages(exclude=[
            'example',
            'example.*',
        ]),
        package_data={
            'tags_input': [
                'templates/*.html',
                'static/js/*.js',
                'static/css/*.css',
                'static/css/base/*.css',
                'static/css/base/images/*.png',
            ],
        },
        extras_require={
            'docs': [
                'django<2',
                'mock',
                'sphinx>=1.7.2',
            ],
            'tests': [
                'django-utils2>=2.5.0',
                'pytest',
                'pytest-cache',
                'pytest-cov',
                'pytest-django',
                'pytest-flakes',
                'pytest-pep8',
            ],
        },
        long_description=long_description,
        cmdclass={'test': PyTest},
        classifiers=['License :: OSI Approved :: BSD License'],
        install_requires=[
            'six',
        ]
    )

